"""
This module provides functions to graphically represent the scattering transform.

Graphviz must be installed.
"""

import numpy as np

import tempfile

from graphviz import Digraph

import matplotlib.pyplot as plt
from matplotlib.figure import Figure 
from matplotlib.backends.backend_svg import FigureCanvasSVG # SVG plotting

from skimage.io import imsave
from skimage.transform import resize

from scatteringtree import scatter_single_node

from itertools import zip_longest


def visualize_tree(scattering_tree, signal, view=True):
    """Display the nodes of a scattering-tree graphically.

    Parameters
    ----------
    
    scattering_tree :   A pre-built scattering_tree object.
    signal          :   The input signal.
    view            :   If set to True then the output will immediately be displayed.
                        Default: True.
    
    Returns:
    A graphviz `graph` object.
    """
    
    st = scattering_tree
    
    return _visualize_tree(signal, st._prop_filters, st._nonlin, st._pools, view=view)


def _save_image(signal, filename):
    """Create a graphical representation of the signal and store it in a file.
    """
    dim =  len(signal.shape)
    
    assert dim in [1,2], "Can plot only 1 and 2 dimensional signals."
    
    signal = np.real(signal)
        
    if dim == 1:
        # 1D plot.
        
        fig = plt.figure()
        ax = fig.add_subplot(111)
        fig.set_size_inches(1,1, forward=True)
        
        x = np.arange(0, len(signal))
        ax.plot(x, signal)
        
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        fig.savefig(filename, bbox_inches='tight', pad_inches=0, dpi=100)
        plt.close()
    elif dim == 2:
        img = signal
        img = resize(img, (64, 64))
        
        img = img - np.min(img)
        img /= np.max(img) # normalize to 0..1
        
        img = img.astype(np.float)
        
        def curve(x,shift):
            return np.exp((-(x-shift)**2)*4)
            #return 1-np.abs(x-shift)
        
        # Generate fancy colors.
        r = curve(img, 1)
        g = curve(img, 0.7)
        b = curve(img, 0.4)
        
        img = np.zeros(img.shape + (3,))
        img[:,:,0] = r
        img[:,:,1] = g
        img[:,:,2] = b
        
        img = img - np.min(img)
        img /= np.max(img) # normalize to 0..1
        img *= 2
        img -= 1
        
        imsave(filename, img)
    
    

def _visualize_tree(signal, propagation_filters, nonlinearities, poolings, view=True):

    # create temporary directory
    tempdir = tempfile.mkdtemp()
    
    
    def create_path(filename, suffix=None):
        """Create a path to a temporary file in a temporary directory.
        """
        if suffix is not None:
            return '%s/%s.%s'%(tempdir, filename, suffix)
        return '%s/%s'%(tempdir, filename)
    
    graph = Digraph(comment='scattering tree', engine='dot')
    
    output = []
    nodes = [signal]
   
        
    meta = [None]*len(nodes)
    
    parent_id = [None]
    
    level = 0
    for prop_filters, nl, pooling in zip_longest(propagation_filters, nonlinearities, poolings):
        next_nodes = []
        next_meta = []
        next_parent_id = []
        
        assert(len(parent_id) == len(nodes))
        assert(len(meta) == len(nodes))
        
        # scatter each node on that level
        for n, m, parent, i in zip(nodes, meta, parent_id, range(0, len(nodes))):
            
            # create node
            node_id = 'node_%d_%d'%(level, i)
            imgfile = create_path(node_id, 'png')
            _save_image(n, imgfile)
            graph.node(node_id, label='', shape='square', image=imgfile)
            
            # draw edge
            if parent is not None:
                graph.edge(parent, node_id)
            
            sc = scatter_single_node(n, prop_filters, nl, pooling, meta=m)
            out = sc['out']
            
            def gen_node_label(node_id, descr):
                html = """<<TABLE BORDER="0">
                                <TR><TD><IMG SRC="%s"/></TD></TR>
                                %s
                            </TABLE>>"""
                
                descr_html = ''
                for k,v in descr.items():
                    descr_html += """
                    <TR><TD>%s</TD><TD>%s</TD></TR>
                    """%(k,v)
                
                return html%(create_path(out_node, 'png'), descr_html)
            
            for o, j in zip(out, range(len(out))):
                
                out_node = 'node_%d_%d_out_%d'%(level, i, j)
                imgfile = create_path(out_node, 'png')
                _save_image(o, imgfile)
                label = gen_node_label(out_node, {'output': '', 'shape': o.shape})
                graph.node(out_node, label=label, color='green', style='dashed', shape='box')
                graph.edge(node_id, out_node, constraint='true')
            
            prop = sc['prop']
            output.extend(out)
            next_nodes.extend(prop)
            next_meta.extend(sc['meta'])
            
            next_parent_id.extend([node_id]*len(prop))
            
        nodes = next_nodes
        meta = next_meta
        parent_id = next_parent_id
        level += 1
    
    if view:
        graph.render(create_path('scatteringtree'), view=True)
    
    return graph
